#!/usr/bin/python
#
# Copyright 2014 Microsoft Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import array
import hashlib
import inspect
import os
import os.path
import stat
import string
import sys
import syslog
from defusedxml.ElementTree import parse
from daemon import runner

import azsec.baselinescan as baselinescan
import azsec.patchscan as patchscan
import azsec.softwarescan as softwarescan
import azsec.scanmanager as scanmanager
import azsec.util as util
from azsec.util import waagent

# Parse command line arguments, set options as appropriate
def ParseArguments():
    parser = argparse.ArgumentParser(description='Scan the system and compare with baseline security standard')
    parser.add_argument(
        '-v', '--verbose', help='write output to stdout', action='count')
    parser.add_argument(
        '-b', '--{0}'.format(baselinescan.scanName), help='perform a baseline configuration scan', action='store_true')
    parser.add_argument(
        '-bc', '--{0}'.format(baselinescan.check), help='restrict baseline scan to the specified check ID (by msid)')
    parser.add_argument(
        '-f', '--file', help='perform scans per the specified config file')
    parser.add_argument(
        '-p', '--{0}'.format(patchscan.scanName), help='perform a patch scan', action='store_true')
    parser.add_argument(
        '-s', '--{0}'.format(softwarescan.scanName), help='perform a software inventory scan', action='store_true')
    parser.add_argument(
        '-start', '--startDaemon', help='start scanningdaemon', action='store_true')
    parser.add_argument(
        '-stop', '--stopDaemon', help='stop scanningdaemon', action='store_true')
    parser.add_argument(
        '-restart', '--restartDaemon', help='restart scanningdaemon', action='store_true')
    args = parser.parse_args()
    # Set verbosity level for writing to console
    if args.verbose:
        util.runtimeVerbosity = args.verbose
    return args

# Try initiating a scan based on options provided
# Return true/false if scan was/was not initiated
# Return num findings from the scans
def TryPerformManualScan(args):
    if not args:
        util.WriteLog(syslog.LOG_ERR, "Invalid object was passed when initiating a scan")
        return False, 0
    if not (args.baselinescan or args.patchscan or args.softwarescan):
        return False, 0
    syslog.openlog(facility=util.defaultFacility)
    scans = args.__dict__
    scanner = scanmanager.scanmanager()
    exitCode = scanner.PerformScan(scans)
    syslog.closelog()
    return True, exitCode

# Try to start/stop/restart the daemon
def TryManageDaemon(args):
    # If invalid param is specified, return
    if (len(sys.argv) < 2):
        return
    # Determine which daemon operation is being attempted and set sys.argv[1] appropriately
    if args.startDaemon:
        sys.argv[1] = "start"
    elif args.stopDaemon:
        sys.argv[1] = "stop"
    elif args.restartDaemon:
        sys.argv[1] = "restart"
    else:
        # No valid daemon options specified
        return
    syslog.openlog(facility=util.defaultFacility)
    util.WriteLog(syslog.LOG_INFO, "Attempting to {0} the daemon".format(sys.argv[1]))
    syslog.closelog()
    try:
        scanner = scanmanager.scanmanager()
        daemonRunner = runner.DaemonRunner(scanner)
        daemonRunner.do_action()
        sys.exit(0)
    except Exception:
        print "An exception occurred while attempting to '{0}' the daemon.".format(sys.argv[1])
        print "Confirm you have appropriate permissions and that the daemon is in the expected state"
        sys.exit(1)

# Parse arguments and config file, kick off scans, report results
def main():
    args = ParseArguments()
    # If parameters indicate we should run a manual scan, do run it
    success, exitCode = TryPerformManualScan(args)
    if success:
        sys.exit(exitCode)
    TryManageDaemon(args)
    # If we got this far, the daemon isn't running
    print "No valid options specified. Run " + __file__ + " -h for help"
    sys.exit(1)

main()
