#!/usr/bin/python
#
# Copyright 2014 Microsoft Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import array
import datetime
import dateutil.parser
import hashlib
import inspect
import isodate
import os
import os.path
import stat
import string
import sys
import syslog
import time
import xml.etree.ElementTree as elementTree

import baselinescan
import patchscan
import softwarescan
import util
from util import waagent
from defusedxml.ElementTree import parse

# Initialize waagent
waagent.LoggerInit('/var/log/waagent.log','/dev/null')

# General Settings
configFile = util.configDir + "/scanner.xml"
lastScanFile = util.configDir + "/lastScan.xml"
epoch = dateutil.parser.parse('1970-01-01 00:00:00')
# Dictionary of functions so we can invoke without hard-coding them
scanFunctions = {}
scanFunctions[baselinescan.scanName] = baselinescan.RunSecureBaselineScan
scanFunctions[patchscan.scanName] = patchscan.RunPatchScan
scanFunctions[softwarescan.scanName] = softwarescan.RunSoftwareInventoryScan
   
class scanmanager():
    scannerConfig = ''
    scanList = [baselinescan.scanName, patchscan.scanName, softwarescan.scanName]
    
    def __init__(self):
        pidPath='/var/run/azsecd'
        self.stdin_path = '/dev/null'
        self.stdout_path = '/dev/null'
        self.stderr_path = '/dev/null'
        self.pidfile_timeout = 3
        if not os.path.exists(pidPath):
            os.makedirs(pidPath)
        self.pidfile_path = "{0}/azsecd.pid".format(pidPath)
    
    def TryScan(self):
        util.WriteLog(syslog.LOG_INFO, "Attempting to perform scheduled scans")
        success, scanSchedule = self.ParseConfig()
        if not success:
            util.WriteConsole(syslog.LOG_ERR, "Could not parse a usable scan schedule from {0}".format(configFile))
            return
        success, lastScan = self.ParseLastScan()
        if not success:
            util.WriteConsole(syslog.LOG_WARNING, "Could not parse last scan times from {0}, will assume this is the first run".format(lastScanFile))
            self.CreateDefaultLastScanFile()
            return
        # Determine which scans to run
        scansToRun = self.DetermineWhichScansToRun(scanSchedule, lastScan)
        # Run Scans
        self.PerformScan(scansToRun)
        # Reset latest scan times for any scans that were run, write to file
        for scanName in self.scanList:
            if (scansToRun[scanName]):
                lastScan[scanName] = datetime.datetime.now()
        self.CreateLastScanFile(lastScan)
        
    def DetermineWhichScansToRun(self, scanSchedule, lastScan):
        # Check for bad input
        scansToRun = {}
        now = datetime.datetime.now()
        for scanName in self.scanList:
            if scanSchedule[scanName] is None:
                scansToRun[scanName] = False
                continue
            elapsed = now - lastScan[scanName]
            scansToRun[scanName] = elapsed >= scanSchedule[scanName]
        return scansToRun
        
    def CreateDefaultLastScanFile(self, outFile = lastScanFile):
        lastScan = {}
        for scanName in self.scanList:
            lastScan[scanName] = epoch
        return self.CreateLastScanFile(lastScan, outFile)
        
    def CreateLastScanFile(self, lastScanTimes, outFile = lastScanFile):
        root = elementTree.Element("lastScan")
        scanners = elementTree.SubElement(root, "scanners")
        for scanName in self.scanList:
            scanners.set(scanName, util.FormatDatetime(lastScanTimes[scanName]))
        tree = elementTree.ElementTree(root)
        tree.write(outFile)
        if os.path.exists(outFile):
            return True
        return False
        
    def ParseLastScan(self, inFile=lastScanFile):
        if not inFile:
            util.WriteLog(syslog.LOG_ERR, "File path for previous scan times must be a valid string")
            return False, None
        if not os.path.isfile(inFile):
            util.WriteLog(syslog.LOG_ERR, "No config file exists at the path" + inFile)
            return False, None
        try:
            elementTree = parse(inFile)
            root = elementTree.getroot()
            scanners = root.find('scanners')
        except:
            util.WriteLog(syslog.LOG_ERR, "Unable to parse file as XML: " +  inFile)
            return False, None
        # Iterate through values, if a value doesn't exist, create it
        lastScanTimes = {}
        for scanName in self.scanList:
            lastScanTimes[scanName] = self.ParseConfigEntryToDateTime(scanners, scanName)
        return True, lastScanTimes

    def ParseConfigEntryToDateTime(self, scanners, scanName):
        if scanners is None or scanName is None:
            # If no valid object specified, parse as Epoch time
            return epoch
        try:
            return dateutil.parser.parse(scanners.get(scanName))
        except:
            util.WriteLog(syslog.LOG_ERR, "Unable to parse a datetime entry from last scan file")
        return epoch
    
    # Return True if a valid duration could be set
    # Return False if not (and set duration to "None")
    def ParseConfigEntryToIso8601(self, scanSchedule, scanners, scanName):
        if scanSchedule is None:
            util.WriteLog(syslog.LOG_ERR, "No scanSchedule object was passed when parsing the scan schedule")
            return False
        if scanners is None:
            util.WriteLog(syslog.LOG_ERR, "No XML data was passed when parsing the scan schedule")
            scanSchedule[scanName] = None
            return False
        duration = scanners.get(scanName)
        # If explicitly set to disabled, silently set to none and return
        if (util.IsDisabled(duration)):
            scanSchedule[scanName] = None
            return False
        success, scanSchedule[scanName] = util.ParseISO8601ToDuration(duration)
        if not success:
            util.WriteLog(syslog.LOG_ERR, "Could not parse schedule for {0}, assuming it is disabled".format(scanName))
        return success
    
    # Return True and a schedule object if at least one scan schedule could be parsed
    # Return False if error or no scans scheduled
    def ParseConfig(self, _scanConfig = configFile):
        # If any problem reading scanner.xml, give up
        if not _scanConfig:
            util.WriteLog(syslog.LOG_ERR, "Config file path must be a valid string")
            return False, None
        if not os.path.isfile(_scanConfig):
            util.WriteLog(syslog.LOG_ERR, "No config file exists at the path" + _scanConfig)
            return False, None
        # If any problems reading 
        self.scannerConfig = _scanConfig
        util.WriteVerbose("Parsing config file " + self.scannerConfig)
        try:
            elementTree = parse(self.scannerConfig)
            root = elementTree.getroot()
            scanners = root.find('scanners')
        except:
            util.WriteLog(syslog.LOG_ERR, "Unable to parse file as XML: " +  self.scannerConfig)
            return False, None
        scanSchedule = {}
        isScanEnabled = False
        for scanName in self.scanList:
            isScanEnabled |= self.ParseConfigEntryToIso8601(scanSchedule, scanners, scanName)
        if not (isScanEnabled):
            util.WriteConsole(syslog.LOG_ERR, "No scans enabled in " + self.scannerConfig)
            return False, None
        return True, scanSchedule

    def PerformScan(self, selectedScans):
        if selectedScans is None:
            util.WriteLog(syslog.LOG_ERR, "Invalid object was passed when initiating a scan")
            return 1
        findings = 0
        attemptedScans = 0
        errCode = 0
        for scanName in self.scanList:
            if (selectedScans[scanName]):
                # Allow customized scans for baseline and AV scanning
                if (scanName is baselinescan.scanName and baselinescan.check in selectedScans):
                    e,f = scanFunctions[scanName](selectedScans[baselinescan.check])
                else:
                    e,f = scanFunctions[scanName]()
                errCode += e
                findings += f
                attemptedScans += 1
        logpri = syslog.LOG_WARNING if errCode else syslog.LOG_INFO
        util.WriteConsole(logpri, "Completed " + `attemptedScans` + " scan(s) with " + `errCode` + " errors and " + `findings` + " findings")
        return errCode
    
    def run(self):
        # If running as a daemon, prevent print statements from spewing to stdout
        sys.stdout = open('/dev/null', 'w')
        while True:
            self.TryScan()
            time.sleep(300)
