#!/usr/bin/python
# Contains utility methods that are available with the default python installation
# These are available to the handler with no modification to the system
#
# Copyright 2014 Microsoft Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import imp
import isodate
import json
import os
import os.path
import platform
import re
import syslog

from defusedxml.ElementTree import parse
from jsonschema import validate

import xml.etree.ElementTree as etree

defaultFacility = syslog.LOG_AUTH
configDir="/etc/azsec"
defaultConfigFile = configDir + "/scanner.xml"
scriptDir=os.path.dirname(os.path.realpath(__file__))
ubuntu = "Ubuntu"
runtimeVerbosity = 0
disabled = "Disabled"
# Scans can be no more frequent than every 5 minutes
minimumDurationSeconds = 300

scanList = ['baselinescan', 'patchscan', 'softwarescan']

def IsConfigMatch(proposedState, constState):
    """ Return true if the string 'constState' starts with 'proposedState' in a case insensitive match """
    if proposedState is None or constState is None:
        return False
    return constState.lower().startswith(proposedState.lower())
    
def IsDisabled(proposedState):
    """Return true if the string 'disabled' starts with the specified string in a case insensitive match"""
    return IsConfigMatch(proposedState, disabled)

def IsValid(proposedState):
    """ If the proposed state is neither Disabled nor a parseable iso8601 duration, return false 
    If the proposed state is "disabled" return True
    Otherwise, return the datetime value represented by this string"""
    if not proposedState:
        return False
    if IsDisabled(proposedState):
        return True
    success, duration = ParseISO8601ToDuration(proposedState)
    return success

def FormatDatetime(dateTime):
    """Turn the provided datetime object into a formatted string"""
    if dateTime is None:
        return ''
    return dateTime.strftime('%Y-%m-%d %H:%M:%S')

def ParseISO8601ToDuration(isoDuration):
    """ Parse the provided string into a time duration"""
    # Fail if not a parseable string
    if not isoDuration:
        return False, None
    try:
        duration = isodate.parse_duration(isoDuration)
        return True, duration
    except:
        return False, None

def searchWAAgent():
    """ Load the waagent and expose it as a submodule """
    agentPath = '/usr/sbin/waagent'
    if(os.path.isfile(agentPath)):
        return agentPath
    user_paths = os.environ['PYTHONPATH'].split(os.pathsep)
    for user_path in user_paths:
        agentPath = os.path.join(user_path, 'waagent')
        if(os.path.isfile(agentPath)):
            return agentPath
    return None

def DistInfo(fullname=0):
    """ Define the function in case waagent(<2.0.4) doesn't have DistInfo() """
    if 'FreeBSD' in platform.system():
        release = re.sub('\-.*\Z', '', str(platform.release()))
        distinfo = ['FreeBSD', release]
        return distinfo
    if os.path.isfile('/etc/oracle-release'):
        release = re.sub('\-.*\Z', '', str(platform.release()))
        distinfo = ['Oracle', release]
        return distinfo
    if 'linux_distribution' in dir(platform):
        distinfo = list(platform.linux_distribution(\
                        full_distribution_name=fullname))
        # remove trailing whitespace in distro name
        distinfo[0] = distinfo[0].strip()
        return distinfo
    else:
        return platform.dist()

def ReadJsonFile(filePath):
    """ Attempt to parse the json file,
    return true if successful, false if not"""
    if not filePath:
        error = "Empty string cannot be parsed as a filepath"
        return False, None
    if not os.path.exists(filePath):
        error = "File {0} does not exist".format(filePath)
        return False, None
    try:
        with open(filePath, 'r') as file:
            result = json.load(file)
            return True, result
    except:
        error = "Failed to parse JSON from {0}".format(filePath)
        return False, None

def ValidateJsonSchema(jsonContent, jsonSchema):
    """ Return true if the provided json content is validated, false if not """
    if not jsonContent or not jsonContent:
        error = "Invalid parameter passed to ValidateJsonSchema"
        return False
    try:
        validate(jsonContent, jsonSchema)
        return True
    except:
        return False

def GetXmlFileElementTree(_configFile):
    """ Open the specified file, parse as xml, and return the element tree object """
    if not _configFile:
        WriteConsole(syslog.LOG_ERR, "Config file path must be a valid string")
        return False, 0
    if not os.path.isfile(_configFile):
        WriteConsole(syslog.LOG_ERR, "No config file exists at the path " + _configFile)
        return False, 0
    # Parse the file
    try:
        WriteVerbose("Parsing config file " + _configFile)
        elementTree = parse(_configFile)
        return True, elementTree
    except:
        WriteConsole(syslog.LOG_ERR, "Unable to parse file as XML: " +  _configFile)
        return False, 0
        
def GetScannerAttribute(_configFile, attribute):
    """ Open the specified file, parse as xml, and retrieve the specified attribute from 'scanners'"""
    success, eTree = GetXmlFileElementTree(_configFile)
    if not (success):
        return False, 0
    scanners = eTree.find('scanners')
    result = scanners.get(attribute)
    if (result is None):
        return False, result
    return True, result
    
def GetSyslogFacility(attribute):
    """ Given an attribute name, parse the default config file return either:
        1. The value from config file (if valid)
        2. The default value (if not valid) """
    success, facility = GetScannerAttribute(defaultConfigFile, attribute)
    if (not success):
        return defaultFacility
    # auth -> LOG_AUTH
    facilityValue = getattr(syslog, "LOG_{0}".format(facility.upper()))
    return facilityValue
    

def ModifyConfig(config, scan, desiredState):
    """ Return true if config is actually modified, false if not """
    if not (scan or desiredState) or config is None:
        WriteConsole(syslog.LOG_ERR, "An invalid parameter was passed while attempting to modify config")
        return False
    WriteDebug("Processing request for " + scan)
    # What happens if elements don't exist
    scanners = config.find('scanners')
    scanState = scanners.get(scan)
    if (IsDisabled(desiredState)):
        desiredState = disabled
    elif (scan in scanList and not IsValid(desiredState)):
        # Scanners should all be ISO8601 durations or 'disabled'
        WriteConsole(syslog.LOG_ERR, "* [{0}] Command not recognized: {1}. Please use a ISO8601 formatted time duration (e.g. 'PT1H') or 'disable'".format(scan, desiredState))
        return False
    WriteDebug("Request to set {0} to {1}".format(scan, desiredState))
    if IsConfigMatch(scanState, desiredState):
        WriteConsole(syslog.LOG_INFO, "* {0} is already {1}".format(scan, desiredState))
        return False
    scanners.attrib[scan] = desiredState
    WriteConsole(syslog.LOG_INFO, "* Set {0} to {1}".format(scan, desiredState))
    return True

def CreateXmlFile(path, tree, uid, gid, mode):
    try:
        fd = open(path, 'a')
        tree.write(fd)
        fd.close()
        os.system('chmod {0} {1}'.format(mode, path))
        os.system('chown {0}:{1} {2}'.format(uid, gid, path))
        return True
    except:
        WriteConsole(syslog.LOG_ERR, "Failed to create " + path)
        return False

def CreateDefaultScannerFile(_configFile):
    """ Creates default scanner config file with all scanners disabled """
    root = etree.Element('config')
    scanners = etree.Element('scanners')
    for scanName in scanList:
        scanners.set(scanName, disabled)
    root.append(scanners)
    return CreateXmlFile(_configFile, etree.ElementTree(root), "root", "root", "644")

def ModifyConfigs(args, _configFile = defaultConfigFile):
    """ Attempts to modify the specified configs. If a change occurs, flush to file """
    if not os.path.isfile(_configFile) and not CreateDefaultScannerFile(_configFile):
        return 1
    success, elementTree = GetXmlFileElementTree(_configFile)
    if not success:
        return 1
    config = elementTree.getroot()
    exitCode = 0
    modified = False
    dictionary = args.__dict__
    for scanName in scanList:
        if(scanName in dictionary and dictionary[scanName] is not None):
            modified |= ModifyConfig(config, scanName, dictionary[scanName])
    if modified:
        WriteLog(syslog.LOG_INFO, "Writing updated config to " + _configFile)
        with open(_configFile, 'w') as ofile:
            elementTree.write(ofile)
    return exitCode

# 0 = Default message verbosity (-v), write to syslog
# 1 = Verbose output (-vv), never write to syslog
# 2 = Debug output (-vvv), never write to syslog
def WriteLog(level, message, messageVerbosity=0):
    if (len(message) < 0):
        return
    if (messageVerbosity < 1):
        syslog.syslog(level, message)
    if (messageVerbosity < runtimeVerbosity):
        print message

# Writes the specified message with a verbosity that is guaranteed to write to console
def WriteConsole(level, message):
    WriteLog(level, message, -1)

# Writes the specified message with verbosity level 1 (-vv)
def WriteVerbose(message):
    WriteLog(syslog.LOG_INFO, message, 1)

# Writes the specified message with verbosity level 2 (-vvv)
def WriteDebug(message):
    WriteLog(syslog.LOG_INFO, message, 2)
    
distro = DistInfo()[0]
agentPath = searchWAAgent()
if(agentPath):
    waagent = imp.load_source('waagent', agentPath)
else:
    raise Exception("Can't load waagent.")
